#  Copyright (c) 2004-5 Aaron S. Lav
#  Permission is hereby granted, free of charge, to any person
#  obtaining a copy of this software and associated documentation
#  files (the "Software"), to deal in the Software without
#  restriction, including without limitation the rights to use, copy,
#  modify, merge, publish, distribute, sublicense, and/or sell copies
#  of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:

#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
#ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.


import sys
import traceback
import urlparse

import gzip
import re
import cStringIO

from twisted.web.proxy import Proxy, ProxyClient, ProxyClientFactory, ProxyRequest
from twisted.internet import reactor


import globalvars
import rewriter
import media_type
import stats


warn_code = 214 # transformation applied
warn_agent = "%s:%d" % (globalvars.web_host, globalvars.proxy_port)
warn_text = 'Annotated by MiddleProxy http://www.pobox.com/~asl2/software/MiddleProxy/'
warnstr = '%d %s "%s"' % (warn_code, warn_agent, warn_text.replace ('"', '""'))

trace_headers = 0
debug_ctr = 0
trace_chunked = 0

# from http://www.xhaus.com/alan/python/httpcomp.html
# It would be nice to avoid the cStringIO stuff and use zlib
# directly, but there's a different header/checksum.  Basically,
# I'd have to hack the gzip.py from the standard library to only
# do the header munging, and present a string interface instead of a file
# one.

def compressBuf(buf):
    zbuf = cStringIO.StringIO()
    zfile = gzip.GzipFile(mode = 'wb',  fileobj = zbuf, compresslevel = 9)
    zfile.write(buf)
    zfile.close()
    return zbuf.getvalue()

def decompressBuf (buf):
    dzfile = gzip.GzipFile(mode = 'rb', fileobj = cStringIO.StringIO(buf))
    return dzfile.read ()

class LFTolerantProxyClient(ProxyClient):
    """Separate lines by LF only, suppress CR if present.  See RFC2616
    19.3 for a description of the suggested algorithm, which I
    implement."""
    delimiter = '\n'
    to_strip = '\r'
    def lineReceived (self, line):
        # XXX optimize!
        if len (line) == 0 or line == '\r':
            return ProxyClient.lineReceived (self, "")
        if line [-1] == self.to_strip:
            line = line[:-1]
        return ProxyClient.lineReceived (self, line)
    
class SecurityMixin(object):
    def securityErr (self, to_write, err):
        to_write.transport.write("HTTP/1.0 403 Forbidden\r\n")

        to_write.transport.write("Content-Type: text/plain\r\n")
        to_write.transport.write("\r\n")
        to_write.transport.write(err)
        to_write.transport.loseConnection()

    
class MyProxyClient (LFTolerantProxyClient, SecurityMixin):
    rewrite_types = [('text', 'html')]
    headers_to_delete_always = [
        'connection', 'keep-alive', 'proxy-authenticate',
        'proxy-authorization', 'te', 'trailers', 
        'upgrade']
    headers_to_delete_rewrite = headers_to_delete_always + [
        'range', 'if-range', 'content-md5', 'transfer-encoding']

    # In theory, transfer-encoding is hop-by-hop, but unless we
    # need to rewrite, we don't need to change it
    
    def __init__ (self, *args, **kw):
        self.rewrite_needed = 0
        self.headers_sent = 0
        self.gzipped = 0
        self.chunked = 0
        self.old_charset = None
        self.buf_list = []
        ProxyClient.__init__ (self, *args, **kw)
        self.to_client_headers = {}
        self.link_str = globalvars.link_str
        self.body_expected = 1
        self.handleResponseEndCalled = 0
        # looks like handleResponseEnd gets called twice sometimes,
        # perhaps when graceful close notification + connection lost notif.

    def sendCommand(self, command, path):
        # XXX make configurable 1.1 vs. 1.0, cache server caps, etc.!
        print command, path
        clower = command.lower ()
        if clower == 'head':
            self.body_expected = 0
        if clower == 'post' and globalvars.restricted:
            err = """command %s path %s unacceptable
to middleproxy in restricted mode.\r\n""" % (command, path)
            return self.securityErr (self.father, err)

        self.transport.write('%s %s HTTP/1.0\r\n' % (command, path))


    def handleStatus (self, version, code, message):
        self.status_version = version
        self.status_code = code
        self.status_message = message
        int_code = int (code)
        if (int_code >= 100 and int_code < 200 or
            int_code == 204 or
            int_code == 304):
            self.body_expected = 0
        

    def handleContentType (self, ct):
        if ct == '':
            return
        mt = media_type.MediaType (ct, self.rest)
        if mt.typ in self.rewrite_types:
            self.rewrite_needed = 1
            self.old_charset = mt.parms.get ('charset', self.old_charset)
            mt.parms ['charset'] = 'UTF-8'
            new = mt.to_inet ()
        else:
            new = ct
            
        self.to_client_headers ['content-type'] = [new]
#        if self.rewrite_needed:
#            print 'Content-Type rewritefl', self.rewrite_needed, repr(new)
        
    def handleContentEncoding (self, ce):
        self.gzipped = ce == 'gzip' or ce == 'x-gzip'

    def sendHeader (self, header, value):
        if trace_headers:
            print "sending", header, value
        h = header.lower ()
        if h in self.headers_to_delete_always:
            if trace_headers:
                print "deleting", header
            return
        if h == 'accept-encoding':
            value = 'identity, gzip' # XXX should handle deflate, compress
        ProxyClient.sendHeader(self, header, value)
        
    def handleHeader (self, key, value):
        "handleHeader is incoming from server"
        key = key.lower ()
        # note: can't delete content-length outbound w/out breaking
        # some POSTs
        if key == 'content-length':
            return # XXX when we support persistent conns, rethink
        if trace_headers:
            print "rcvd header", key, value
        if self.to_client_headers.has_key (key):
            self.to_client_headers [key].append (value)
            return
        if key == 'server':
            stats.inc_stats (stats.STAT_SERVER, value)
        # XXX what if content-type, content-encoding repeated?
        if key == 'content-type':
            self.handleContentType (value)
        else:
            if key == 'content-encoding':
                self.handleContentEncoding (value)
            elif key == 'transfer-encoding':
                self.chunked = value == 'chunked'
            self.to_client_headers [key] = [value]

    def handleEndHeaders (self):
        if not self.body_expected:
            self.sendHeaders ()
            self.handleResponseEnd ()


    def sendEndHeaders (self):
        if self.rewrite_needed:
            to_del = self.headers_to_delete_rewrite
        else:
            to_del = self.headers_to_delete_always

        for h in to_del:
            if self.to_client_headers.has_key (h) and self.rewrite_needed:
                del self.to_client_headers [h]
        for k, v_list in self.to_client_headers.items ():
            for v in v_list:
                ProxyClient.handleHeader (self, k, v)
            
        ProxyClient.handleEndHeaders (self)

    def sendHeaders (self):
        ProxyClient.handleStatus (self,
                                  self.status_version,
                                  self.status_code,
                                  self.status_message)
        self.sendEndHeaders ()
        self.headers_sent = 1
        
    def handleResponsePart (self, buffer):
        if self.rewrite_needed:
            self.buf_list.append (buffer)
        else:
            if not self.headers_sent:
                self.sendHeaders ()

            ProxyClient.handleResponsePart (self, buffer)
            
    trace_rewrite = 0
    http_equiv_re = re.compile(
        '<META\s+HTTP-EQUIV="?Content-Type"?\s+content="(.*?)"\s?/?\s?>',
        re.IGNORECASE)
    # Note that it's legal to omit quotes from http-equiv value, but
    # not content, since content contains a '/'.  See HTML4.0 3.2.2.
    
    def extract_charset (self):

        mo = self.http_equiv_re.search (self.buf)
        if mo <> None:
            content = mo.group (1)
            mt = media_type.MediaType (content, self.rest)
            new_charset = mt.parms.get ('charset', None)

            if new_charset and new_charset.lower () <> 'utf-16': 
                self.old_charset = new_charset

            # Note there's a sometimes a META encoding declaration w/ UTF-16
            # when the original 'META ...' text is in some 8-bit charset.
            # I'm not sure if this ever happens without a charset
            # declaration in the content-type header, but such a
            # a declaration can't possibly be correct, since the
            # META declaration itself is 8-bits/char

            # See the following links for other reports of the same problem:
            # http://lists.evolt.org/archive/Week-of-Mon-20020722/119196.html
            # http://www.fuzzysoftware.com/newsgroups.asp?action=showmsgs&group=microsoft.public.xml&messid=84072
            

            self.buf = self.buf [0:mo.start (0)] + self.buf [mo.end(0):]

        if self.old_charset == None:
            self.old_charset = 'iso-8859-1' # default charset
        else:
            if self.old_charset[0] == '"' and self.old_charset[-1] == '"':
                self.old_charset = self.old_charset [1:-1]
                # www.haaretz.co.il wrongly (?) generates quoted encodings

            # Python 2.2 doesn't know about csISOLatinHebrew, 2.3 and later do
            if self.old_charset.lower () == "csISOLatinHebrew".lower ():
                if trace_headers:
                    print "changing to iso-8859-8"
                self.old_charset = 'iso-8859-8'
            if self.old_charset.lower ().startswith ('gb_2312'):
                print 'truncating gb_2312', self.old_charset
                # e.g. gb_2312-80, which I presume is the 1980 rev
                # of gb2312, and is a subset of the current gb2312
                self.old_charset = 'gb2312'
                # not a typo, python2.4 recognizes gb2312 but not gb_2312

    def rewrite_buf (self):
        # In general, header should override META, but we assume that
        # iso-8859-1 might be a default: get
        # http://schiller.dartmouth.edu/chinese/Texts/Proverbs/5gb.html
        # right.
        if self.old_charset == None or self.old_charset.lower () == 'iso-8859-1':
            self.extract_charset ()
                        
        print "transcoding from", self.old_charset.lower (), "to utf-8 for", self.rest, self.headers.get ('host', 'no host')
        stats.inc_stats (stats.STAT_CHARSETS, self.old_charset.lower ())
        self.buf = self.buf.decode (self.old_charset, 'ignore')
        new_buf = rewriter.rewrite (
            self.buf, self.link_str).encode ('utf-8')
        if self.trace_rewrite:
            global debug_ctr
            open ('/tmp/rewri' + str (debug_ctr) + '.html', 'w').write (self.buf.encode ('utf-8'))
            open ('/tmp/rewro' + str (debug_ctr) + '.html', 'w').write (new_buf)
            debug_ctr += 1
        self.buf = new_buf

    def decode_chunked (self):
        new_buf = ''
        old_buf = self.buf
        crlf = chr(13) + chr (10)
        while 1:
            i = old_buf.find (crlf)
            if trace_chunked:
                print "i", i
            if i == -1:
                print "bad i", repr(old_buf[0:100])
                return
            # have observed transfer-encoding chunked, but data is just plain gzipped (and correctly content-encoding gzip) from www.msnbc.msn.com
            # maybe should only bail like this if no chunks found at all?
            sz = int (old_buf [0:i], 16)
            if trace_chunked:
                print "sz", sz
            if sz == 0:
                break
            new_buf += old_buf [i + len (crlf): i + len (crlf) + sz]
            old_buf =  old_buf[i + len (crlf) + len (crlf) + sz:]
        self.buf = new_buf

    def decode (self):
        if self.chunked:
            self.decode_chunked ()
        if self.gzipped:
            try:
                if len (self.buf) <= 10:
                    self.buf = ""
# I have observed Apache (1.3.31 generate content-encoding
# x-gzip, response code 500, with an entity-body of just the 10-byte
# gzip header, and no data, no CRC.
                else:
                    self.buf = decompressBuf (self.buf)
            except IOError, err:
                print "IO err", err
                print "buf", repr(self.buf[0:100])
        self.rewrite_buf ()
        if self.gzipped:
            self.buf = compressBuf (self.buf)
    errHdrs = [('content-type', 'text/html')]
    def sendErr (self, text_list):
        print "sendErr", text_list
        ProxyClient.handleStatus (self,
                                  self.status_version,
                                  self.status_code, self.status_message)
        for header, value in self.errHdrs:
            ProxyClient.sendHeader(self, header, value)
        ProxyClient.handleEndHeaders (self)
        self.father.transport.write (
            '<html><head><title>Proxy Error</title></head><body><p>')
        self.father.transport.write ("<br>\n".join (text_list))
        self.father.transport.write ('</p></body></html>')
        ProxyClient.handleResponseEnd (self)
            
    def handleResponseEnd (self):
        if not self.rewrite_needed:
            ProxyClient.handleResponseEnd (self)
            return

        if self.handleResponseEndCalled:
            return
        self.handleResponseEndCalled = 1
        try:
            if self.rewrite_needed:
                self.buf = "".join (self.buf_list)
                if len (self.buf) <> 0:
                    self.decode ()
        except:
            print "exn"
            typ, val, tb = sys.exc_info ()
            tb_list = [str(typ), str (val)] +  traceback.format_tb (tb)
            self.status_code = '500'
            self.status_message = 'Internal Proxy Exception'
            return self.sendErr (tb_list) 
            

        ProxyClient.handleStatus (self,
                                  self.status_version,
                                  self.status_code, self.status_message)
        self.sendEndHeaders ()

        if self.rewrite_needed:
            self.father.transport.write (self.buf)
            
        ProxyClient.handleResponseEnd (self)

class MyProxyClientFactory (ProxyClientFactory):
    def buildProtocol(self, addr):
        return MyProxyClient(self.command, self.rest, self.version,
                           self.headers, self.data, self.father)
    def clientConnectionFailed (self, connector, reason):
        ProxyClientFactory.clientConnectionFailed (self, connector, reason)
        self.father.transport.write ('<p>' + str (reason))
        self.father.transport.loseConnection ()
        
class MyProxyRequest (ProxyRequest, SecurityMixin):
    protocols = {'http': MyProxyClientFactory}
    
    
    def process(self):
        """Copied wholesale from twisted, since hooks not present
        for what I want to do."""
        parsed = urlparse.urlparse(self.uri)
        protocol = parsed[0]
        host = parsed[1]
        try:
            port = self.ports[protocol]
        except KeyError:
            # spotted this exn in production XXX
            print "bad protocol", protocol, "for uri", repr (self.uri)
            raise

        if ':' in host:
            host, port = host.split(':')
            port = int(port)
            if port <> 80 and globalvars.restricted:
                return self.securityErr (self, "port %d not allowed in middleproxy restricted mode" % port)
                return

        rest = urlparse.urlunparse(('','')+parsed[2:])
        if not rest:
            rest = rest+'/'
        class_ = self.protocols[protocol]
        headers = self.getAllHeaders().copy()
        if not headers.has_key('host'):
            headers['host'] = host
        self.content.seek(0, 0)
        s = self.content.read()
        clientFactory = class_(self.method, rest, self.clientproto, headers,
                               s, self)
        reactor.connectTCP(host, port, clientFactory)


class MyProxy(Proxy):
    requestFactory = MyProxyRequest

